function [f, e] = fv(B, EigA, nk, thmax, noplot)
%       Field of values (or numerical range).
%       FV(A, NK, THMAX) evaluates and plots the field of values of the NK
%       largest leading principal submatrices of A, using THMAX equally spaced
%       angles in the complex plane.  The defaults are NK = 1 and THMAX = 16.
%       (For a `publication quality' picture, set THMAX higher, say 32.)
%       The eigenvalues of A are displayed as `x'.
%       Alternative usage: [F, E] = FV(A, NK, THMAX, 1) suppresses the plot
%       and returns the field of values plot data in F, with A's eigenvalues
%       in E.   Note that NORM(F,INF) approximates the numerical radius,
%       max {abs(z): z is in the field of values of A}.

%       Further modified starting Oct. 3, 1994 to compute the bounding rectangle.
%       B is the symbol in input because A is used as a principal submatrix.
%       (note as of 10-03-94).
 
%       Theory:
%       Field of values FV(A) = set of all Rayleigh quotients. FV(A) is a
%       convex set containing the eigenvalues of A.  When A is normal FV(A) is
%       the convex hull of the eigenvalues of A (but not vice versa).
%               z = x'Ax/(x'x),  z' = x'A'x/(x'x)
%               => REAL(z) = x'Hx/(x'x),   H = (A+A')/2
%       so      MIN(EIG(H)) <= REAL(z) <= MAX(EIG(H))
%       with equality for x = corresponding eigenvectors of H.  For these x,
%       RQ(A,x) is on the boundary of FV(A).

%       References:
%       R.A. Horn and C.R. Johnson, Topics in Matrix Analysis, Cambridge
%            University Press, 1991, section 1.5.
%       A.S. Householder, The Theory of Matrices in Numerical Analysis,
%            Blaisdell, New York, 1964, section 3.3.
%       C.R. Johnson, Numerical determination of the field of values of a
%            general complex matrix, SIAM J. Numer. Anal., 15 (1978),
%            pp. 595-602.

%       Original version by A. Ruhe.  Modified by N.J. Higham.

if nargin < 3, nk = 1; end
if nargin < 4, thmax = 16; end

iu = sqrt(-1);
[n, p] = size(B);
f = [];
z = zeros(2*thmax+1,1);
% Prevent eigs from showing intermediate output
opts.disp = 0;
% Prevent eigs from showing intermediate output
opts.tol = 1e-1;
e = EigA;
%
% Filter out cases where B is Hermitian or skew-Hermitian, for efficiency.
if norm((B - B')/2.0,1) == 0

   f = [min(e) max(e)];             

elseif norm((B+B')/2,1) == 0

   e = imag(e);
   f = [min(e) max(e)];
   e = iu*e; f = iu*f;
   
else

for m = 1:nk

   ns = n+1-m;
   A = B(1:ns, 1:ns);

   for i = 0:thmax
      th = i/thmax*pi;
      Ath = exp(iu*th)*A;               % Rotate A through angle th.
      H = 0.5*(Ath + Ath');             % Hermitian part of rotated A.
      [X, D] = eigs(H,2,'BE',opts);
%     4/19/01
%		Using the eigs routine to pick off the largest/smallest eigenvalues
%		and corresponding eigenvalues
%
%     RQ's of A corr. to eigenvalues of H with smallest/largest real part.: 
%		smallest part
		z(1+i) = X(:,1)' * A * X(:,1)/(X(:,1)' * X(:,1)); 
%		largest part      
      z(1+i+thmax) = X(:,2)' * A * X(:,2)/(X(:,2)' * X(:,2));
%
   end

   f = [f; z];

end
%
% Change made to compute the bounding rectangle. A is assumed non-Hermitian
% and non-skew-Hermitian. 10-03-94.
%
  i_imag = sqrt(-1.0);
  H_A = 0.5 * (A + A');
  K_A = 0.5 * (A - A')/i_imag;
  egvls_H = real(eigs(H_A,2,'BE',opts));
  egvls_K = real(eigs(K_A,2,'BE',opts));
  min_H   = min(egvls_H);
  min_K   = min(egvls_K);
  max_H   = max(egvls_H);
  max_K   = max(egvls_K);


end

if nargin < 5
   % start with hold off in case of graphics on the screen. 10-4-94. 
   hold off
   % Set x and y axis ranges so both have the same length.
   xmin = min(real(f)); xmax = max(real(f));
   ymin = min(imag(f)); ymax = max(imag(f));
   if xmax-xmin >= ymax-ymin
      ymid = (ymin + ymax)/2;
      ymin =  ymid - (xmax-xmin)/2; ymax = ymid + (xmax-xmin)/2;
   else
      xmid = (xmin + xmax)/2;
      xmin = xmid - (ymax-ymin)/2; xmax = xmid + (ymax-ymin)/2;
   end
   axis('square')
   % Scale ranges by 1+2*alpha to give extra space around edges of plot.
   alpha = 0.1;
   axis( [xmin-alpha*(xmax-xmin) xmax+alpha*(xmax-xmin) ...
          ymin-alpha*(ymax-ymin) ymax+alpha*(ymax-ymin) ] );

   plot(real(f), imag(f),'c')      % Plot the field of values
   hold on
   plot(real(e), imag(e), 'xg')    % Plot the eigenvalues too.
%   axis([1 2 3 4]); axis; axis('normal');  % Normal ratio and auto-scaling.
%
% The first arg of plot is the x-values and the second the y. The vertices
% of the rectangle start with p1 at the NE corner and continue
% counter-clockwise around the rectangle.  
%
%  p1 = [max_H;max_K]       p2          p1
%  p2 = [min_H;max_K]
%  p3 = [min_H;min_K]
%  p4 = [max_H;min_K]       p3          p4
 
% The x-values for p1 and p2. 

   X = [min_H; max_H];
 
% The y-vals for p1 and p2. 
 
  Y = [max_K; max_K];
  plot(X, Y)
  hold on

% X and Y -vals for p2 and p3.

  X = [min_H; min_H]; 
  Y = [min_K; max_K];
  plot(X, Y)

  hold on

% X and Y -vals for p3 and p4. 

  X = [min_H; max_H]; 
  Y = [min_K; min_K];
  plot(X, Y)

  hold on

% X and Y -vals for p4 and p1. 

  X = [max_H; max_H]; 
  Y = [min_K; max_K];
  plot(X, Y)
  hold on
% This completes the section on the bounding rectangle. 10-4-94. 

% Next we try to put in the convex hull of the eigenvalues. 
  EigA = e; 
  ch = convex_hull(EigA);
  if length(ch) > 0, ch = [ch ; ch(1)]; end;
  plot(real(ch), imag(ch), '-r'); 
 
  hold off

end
