/*
          This file is part of Sandia National Laboratories
          copyrighted software.  You are legally liable for any
          unauthorized use of this software.

          NOTICE:  The United States Government has granted for
          itself and others acting on its behalf a paid-up,
          nonexclusive, irrevocable worldwide license in this
          data to reproduce, prepare derivative works, and
          perform publicly and display publicly.  Beginning five
          (5) years after June 5, 1997, the United States
          Government is granted for itself and others acting on
          its behalf a paid-up, nonexclusive, irrevocable
          worldwide license in this data to reproduce, prepare
          derivative works, distribute copies to the public,
          perform publicly and display publicly, and to permit
          others to do so.

          NEITHER THE UNITED STATES GOVERNMENT, NOR THE UNITED
          STATES DEPARTMENT OF ENERGY, NOR SANDIA CORPORATION,
          NOR ANY OF THEIR EMPLOYEES, MAKES ANY WARRANTY, EXPRESS
          OR IMPLIED, OR ASSUMES ANY LEGAL LIABILITY OR
          RESPONSIBILITY FOR THE ACCURACY, COMPLETENESS, OR
          USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT, OR
          PROCESS DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT
          INFRINGE PRIVATELY OWNED RIGHTS.
*/

// Driver Version 5.1.4 & 5.2.0 switchable
// Date: October 1, 1996

// comment out this definition line for v-5.1 matrix classes (non-resizeable)
//#define RESIZEABLE_CRS_MATRIX

#ifdef RESIZEABLE_CRS_MATRIX
    #define SCRS_Matrix RsSCRS_Matrix
#endif
  
//#define ISIS_DEBUG      // comment this out to disable debug output

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <new.h>

#undef NDEBUG // make sure asserts are enabled (driver only)
#include <assert.h>

#include "isis-ser.h"

// Prototypes ...............................

extern "C" double get_cpu_time();
extern "C" double get_wall_clock_time();

static void fillRowCount(int **rowCount, int& n);
static void fillRowCount_bin(int **rowCount, int& n);

static void get_command_line_options(int argc, char **argv);
static int getcfa(const char *flag, int argc, char **argv, char *argstr);

bool binaryData();
bool readIntParam(char *string, int& iparam);
bool readBoolParam(char *string, bool& bparam);
bool readDoubleParam(char *string, double& dparam);
static short ASCII_DATA = 0;

#define MaxStringLength 256

//global variables for filenames
static char Afilename[MaxStringLength] = {"matrix_test"};
static char bfilename[MaxStringLength] = {"rhs_test"};
static char solvername[MaxStringLength] = {"qmr"};
static char pcname[MaxStringLength] = {"identity"};

void noMoreMemory()
{
   cerr << "Unable to allocate more memory.\n" << endl;
   exit(0);
}  

/**=========================================================================**/
int main(int argc, char **argv) {
        
    get_command_line_options(argc, argv);

//  Dummy variables for temp use.

    int idummy; double dummy; bool bdummy;

//  Create dummy pointers for deletion of composed preconditioners.
    Preconditioner* pPC1 = 0;
    Preconditioner* pPC2 = 0;

    int i;

//  Set up timers.
    double solv_time = 0, solv_time0;
    double pc_time = 0, pc_time0;

    set_new_handler(noMoreMemory);

    if(binaryData()){
        cout << "Binary data files" << endl;
        ASCII_DATA = 0;
    }
    else {
        cout << "ASCII data files" << endl;
        ASCII_DATA = 1;
    }

//  Indicate which type of matrix is used.
#ifdef RESIZEABLE_CRS_MATRIX
    cout << "main: v-5.2 resizeable matrices used" << endl;
#else
    cout << "main: v-5.1 matrices used" << endl;
#endif
    
    int **rowCount = new int*;
    *rowCount = NULL;
    int n;
    
    cout << "Getting rowCount information from file..." << endl;

    if (ASCII_DATA==1)
        fillRowCount(rowCount, n);
    else
        fillRowCount_bin(rowCount, n);

//  Construct map -- simple constructor for serial case.
    Map map(n);

// Construct an integer vector, fill it with the row lengths in rowCount, then
// use it to configure the matrix A.
    Seq_IntVector *rowLengths = new Seq_IntVector(map);

    for(i=1; i<=n; i++) (*rowLengths)[i] = (*rowCount)[i-1];

//  Construct an "empty" matrix then inform it of its structure.
    SCRS_Matrix A(map);
    A.configure(*rowLengths);
    
// We don't need rowLenghts or the rowCount array any more.
    if (*rowCount != NULL) delete [] *rowCount;
    delete rowCount;
    delete rowLengths;

    cout << "Reading matrix from data file..." << endl;

    A.readFromFile(Afilename);
    A.fillComplete();

//  Declare solution and RHS vectors

    Seq_Vector x(map);
    x.put(0.0);
    Seq_Vector b(map);

// now proceed with setting up the linear system to solve.
    x.put(0.0);
    b.put(0.0);   // to insure proper initialization

    cout << "Reading rhs vector b from data file..." << endl;
    cout << "I think bfilename is " << bfilename << endl;
    b.readFromFile(bfilename);
    
//  Now select the solver, according to the command-line parameter solvername.

    IterativeSolver *psolver;

    cout << endl << "*****************************" << endl;
    if (!strcmp(solvername, "qmr")) {
        psolver = new QMR_Solver;
        cout << "QMR solver selected." << endl;
    }
    else if (!strcmp(solvername, "qmr2")) {
        psolver = new QMR2_Solver;
        cout << "QMR2 solver selected." << endl;
    }
    else if (!strcmp(solvername, "gmres")) {
        int restart;
        if(!readIntParam("restart",restart)) restart = 90;
        psolver = new GMRES_Solver(restart);
        cout << "GMRES(" << restart << ") solver selected." << endl;
    }
    else if (!strcmp(solvername, "fgmres")) {
        int restart;
        if(!readIntParam("restart",restart))restart = 90;
        psolver = new FGMRES_Solver(restart);
        cout << "FGMRES(" << restart << ") solver selected." << endl;
    }
    else if (!strcmp(solvername, "cg")) {
        psolver = new CG_Solver;
        cout << "CG solver selected." << endl;
    }
    else if (!strcmp(solvername, "defgmres")) {
        int restart;
        if(!readIntParam("restart",restart))restart = 90;
        psolver = new DefGMRES_Solver(restart);
        cout << "DefGMRES(" << restart << ") solver selected." << endl;
    }
    else if (!strcmp(solvername, "bicgstab")) {
        psolver = new BiCGStab_Solver;
        cout << "BiCGStab solver selected." << endl;
    }
    else if (!strcmp(solvername, "cgs")) {
        psolver = new CGS_Solver;
        cout << "CGS solver selected." << endl;
    } 
    else if (!strcmp(solvername, "cgne")) {
        psolver = new CGNE_Solver;
        cout << "CGNE solver selected." << endl;
    } 
    else if (!strcmp(solvername, "cgnr")) {
        psolver = new CGNR_Solver;
        cout << "CGNR solver selected." << endl;
    } 
    else {
        psolver = new QMR_Solver;
        cout << "Defaulting to QMR solver." << endl;
    }

//  Now select the preconditioner.
//  Many preconditioners must be passed a sample vector of the type that
//  they are to use. Any vector of the correct type will work as they use
//  the newVector member function on the sample vector to create a vector
//  to hold the preconditioner.

//  Pointers to base class -- stay in scope.
    Preconditioner *pprecond,*ppc1,*ppc2;

//  Parse input to select preconditioner.
    if (!strcmp(pcname, "identity")) {

        pprecond = new Identity_PC(A);
        cout << "Identity pc selected." << endl;
    }
    else if (!strcmp(pcname, "diagonal")) {

        pprecond = new Diagonal_PC(A, x);
        cout << "Diagonal pc selected." << endl;
    }
    else if (!strcmp(pcname, "polynomial")) {

        Poly_PC* ppc = new Poly_PC(A, x);

        // Set polynomial pc type -- default is type 2 => least squares.
        if (!readIntParam("polyType",idummy)) idummy = 2;
        ppc->polyType(idummy);

        // Set polynomial pc order -- default is order 2.
        if (!readIntParam("polyOrder",idummy)) idummy = 2;
        ppc->polyOrder(idummy);

        cout << "Polynomial pc selected." << endl;
        cout << "- order:     " << ppc->polyOrder() << endl;
        cout << "- type:      " << ppc->polyType() << endl;

        // Cast back to base class.
        pprecond = (Preconditioner*) ppc;
    }
    else if (!strcmp(pcname, "spai")) {
        pprecond = new SPAI_PC(A);
        cout << "SPAI pc selected." << endl;
    }
    else if (!strcmp(pcname, "polybj")) {

        BlockJacobi_PC* ppc = new BlockJacobi_PC(A);
        pPC1 = (Preconditioner*) ppc;  // record pointer for later deletion

        // Set block size.
        if (map.globalSize() < 64) {
            ppc->blockSize(min(map.globalSize(),4));
        }
        else {
            if (!readIntParam("blockSize",idummy)) idummy = 32;
            ppc->blockSize(idummy);
        }

        // Set overlap strategy -- 1=>non-overlap, 2=>overlap.
        if (!readIntParam("blockType",idummy)) idummy = 2; // 2 => overlap
        ppc->blockType(idummy);

        cout << "Block Jacobi pc selected." << endl;
        cout << "- block size: " << ppc->blockSize() << endl;
        cout << "- strategy:   " << ppc->blockType() << endl;

        // Cast back to base class.
        ppc1 = (Preconditioner*) ppc;

        Poly_PC* pppc = new Poly_PC(A, x, *ppc1);
        pPC2 = (Preconditioner*) pppc;  // record pointer for later deletion

        // Set polynomial pc type -- default is type 2 => least squares.
        if (!readIntParam("polyType",idummy)) idummy = 2;
        pppc->polyType(idummy);

        // Set polynomial pc order -- default is order 2.
        if (!readIntParam("polyOrder",idummy)) idummy = 2;
        pppc->polyOrder(idummy);

        cout << "Poly_PC parameters:" << endl;
        cout << "- order:     " << pppc->polyOrder() << endl;
        cout << "- type:      " << pppc->polyType() << endl;

        // Cast back to base class.
        ppc2 = (Preconditioner*) pppc;
        pprecond = new Composed_PC(A,x,*ppc1,*ppc2);

    }
    else if (!strcmp(pcname, "bj")) {

        BlockJacobi_PC* ppc = new BlockJacobi_PC(A);

        // Set block size.
        if (map.globalSize() < 64) {
            ppc->blockSize(min(map.globalSize(),4));
        }
        else {
            if (!readIntParam("blockSize",idummy)) idummy = 32;
            ppc->blockSize(idummy);
        }

        // Set overlap strategy -- 1=>non-overlap, 2=>overlap.
        if (!readIntParam("blockType",idummy)) idummy = 2; // 2 => overlap
        ppc->blockType(idummy);

        cout << "Block Jacobi pc selected." << endl;
        cout << "- block size: " << ppc->blockSize() << endl;
        cout << "- strategy:   " << ppc->blockType() << endl;

        // Cast back to base class.
        pprecond = (Preconditioner*) ppc;
    }
    else if (!strcmp(pcname, "cgne_bj")) {

        CGNE_BlockJacobi_PC* ppc = new CGNE_BlockJacobi_PC(A);

        // Set block size.
        if (map.globalSize() < 64) {
            ppc->blockSize(min(map.globalSize(),4));
        }
        else {
            if (!readIntParam("blockSize",idummy)) idummy = 32;
            ppc->blockSize(idummy);
        }

        // Set overlap strategy -- 1=>non-overlap, 2=>overlap.
        if (!readIntParam("blockType",idummy)) idummy = 2; // 2 => overlap
        ppc->blockType(idummy);

        cout << "CGNE Block Jacobi pc selected." << endl;
        cout << "- block size: " << ppc->blockSize() << endl;
        cout << "- strategy:   " << ppc->blockType() << endl;

        // Cast back to base class.
        pprecond = (Preconditioner*) ppc;
    }
    else if (!strcmp(pcname, "cgne_diag")) {

        pprecond = new CGNE_Diagonal_PC(A, x);
        cout << "CGNE Diagonal pc selected." << endl;
    }
    else if (!strcmp(pcname, "cgne_poly")) {

        CGNE_Poly_PC* ppc = new CGNE_Poly_PC(A, x);

        // Set polynomial pc type -- default is type 2 => least squares.
        if (!readIntParam("polyType",idummy)) idummy = 2;
        ppc->polyType(idummy);

        // Set polynomial pc order -- default is order 2.
        if (!readIntParam("polyOrder",idummy)) idummy = 2;
        ppc->polyOrder(idummy);

        cout << "CGNE Polynomial pc selected." << endl;
        cout << "- order:     " << ppc->polyOrder() << endl;
        cout << "- type:      " << ppc->polyType() << endl;

        // Cast back to base class.
        pprecond = (Preconditioner*) ppc;
    }
    else if (!strcmp(pcname, "cgnr_poly")) {

        CGNR_Poly_PC* ppc = new CGNR_Poly_PC(A, x);

        // Set polynomial pc type -- default is type 2 => least squares.
        if (!readIntParam("polyType",idummy)) idummy = 2;
        ppc->polyType(idummy);

        // Set polynomial pc order -- default is order 2.
        if (!readIntParam("polyOrder",idummy)) idummy = 2;
        ppc->polyOrder(idummy);

        cout << "CGNR Polynomial pc selected." << endl;
        cout << "- order:     " << ppc->polyOrder() << endl;
        cout << "- type:      " << ppc->polyType() << endl;

        // Cast back to base class.
        pprecond = (Preconditioner*) ppc;
    }
    else {

        pprecond = new Identity_PC(A);
        cout << "Defaulting to Identity pc." << endl;
    }
    cout << "*****************************" << endl << endl;

//  Now set solution parameters.

    // Set max iterations.
    if (!readIntParam("maxIterations",idummy)) idummy = 8000;
    psolver->maxIterations(idummy);
    cout << "Max iterations = " << psolver->maxIterations() << endl;

    // Set convergence tolerance.
    if (!readDoubleParam("tolerance",dummy)) dummy = 1.e-10;
    psolver->tolerance(dummy);
    cout << "Convergence tolerance = " << psolver->tolerance() << endl;

//  Construct the system of equations object.

    LinearEquations lse(A, x, b);

//  Column scaling.

    if (!readBoolParam("colScale",bdummy)) bdummy = true;
    if (bdummy) {
       cout << "Column scaling turned on..." << endl;
       if (!lse.colScale()) {
          cout << "Warning - column scaling failed." << endl;
       }
       else {
          cout << "...column scaling completed." << endl;
       }
    }
    else {
       cout << "Column scaling turned off." << endl;
    }

//  Row scaling.

    if (!readBoolParam("rowScale",bdummy)) bdummy = true;
    if (bdummy) {
       cout << "Row scaling turned on..." << endl;
       if (!lse.rowScale()) {
          cout << "Warning - row scaling failed." << endl;
       }
       else {
          cout << "...row scaling completed." << endl;
       }
    }
    else {
       cout << "Row scaling turned off." << endl;
    }

//  Set solver & preconditioner for lse.

    lse.setSolver(*psolver);
    lse.setPreconditioner(*pprecond);
    
    cout << "Computing the preconditioner... " << endl << flush;

    char* pstr = new char[64];
    sprintf(pstr, "stoppingTest Resid_RHS");
    psolver->parameters(1, &pstr);
    delete [] pstr;

    psolver->outputLevel(1,0,0);
    pc_time0 = get_wall_clock_time();
    pprecond->calculate();
    pc_time = get_wall_clock_time() - pc_time0;

    cout << "...finished computing preconditioner." << endl << flush;

    cout << "Starting solve... " << endl << flush;

    solv_time0 = get_wall_clock_time();
    int solveStatus = lse.solve();            
    solv_time = get_wall_clock_time() - solv_time0;

    cout << "...solve complete. " << endl << flush;

    cout << "solver return status " << solveStatus << endl << flush;
    cout << endl;

    delete psolver;

    // check for composed pc's ... delete individual pc's in addition to 
    // composed pc.
    if (pPC1 != 0)
      delete pPC1;

    if (pPC2 != 0)
      delete pPC2;

    delete pprecond;

    cout << endl << "PC calculate wall clock time:   " << pc_time <<
        " sec." << endl;

    cout << "Solve wall clock time:          " << solv_time << " sec." << endl;

    cout << "Solve & PC calc time:           " << pc_time + solv_time <<
        " sec." << endl;
 
#ifdef PRODUCTION
    cout << endl << "Production matrix and vector classes" << endl;
#else
    cout << "Development matrix and vector classes" << endl;
#endif
    
    return(0);
}


/**=========================================================================**/
bool binaryData(){
//
//This simple function checks to see if the file name contains '.mtx'.
//If so, return false (which means the file is NOT binary data).
//Else return true, which means it is binary data.
//
    if (strstr(Afilename, ".mtx") != NULL) {
        return(false);
    }
    else return(true);
}
 
/**=========================================================================**/
void fillRowCount(int **rowCount, int& n) {

/** fillRowCount --- scan all i,j values for "a" to determine row lengths **/

    char line[128];

    FILE *fStrmA = fopen(Afilename,"r");
    if(!fStrmA){
        cout << "Error opening " << Afilename << endl;
        return;
    }

    int i, j;
    double value;

    do {
        fgets(line,128,fStrmA);
    } while(strchr(line,'%'));

    sscanf(line,"%d",&n);
    cout << "in fillRowCount, n=" << n << endl;

    if (*rowCount != NULL) delete [] *rowCount;
    *rowCount = new int[n];

    for(i=0; i<n; i++) (*rowCount)[i] = 0;

    while (!feof(fStrmA))  {
        do {
            fgets(line,128,fStrmA);
        } while(strchr(line, '%'));

        if(feof(fStrmA))return;
        sscanf(line,"%d %d %lg",&i,&j,&value);
        (*rowCount)[i-1]++;  // increment rowCount for row i
    }
    fclose(fStrmA);
    return;
}


/**=========================================================================**/
void fillRowCount_bin(int **rowCount, int& n) {
/** fillRowCount_bin --- (binary) scan all i,j values for "a" to determine 
    row lengths **/

    FILE *fStrmA = fopen(Afilename,"r");
    
    int m, nnz;

    fread((char *)&n,sizeof(int),1,fStrmA);
    fread((char *)&m,sizeof(int),1,fStrmA);
    fread((char *)&nnz,sizeof(int),1,fStrmA);
    cout << "in fillRowCount_bin, read binary n: " << n << ", m: " << m
         << ", nnz: " << nnz << endl;

    if (*rowCount != NULL) delete [] *rowCount;
    *rowCount = new int[n];

    for(int i=0; i<n; i++) (*rowCount)[i] = 0;
    
    struct COEF { int i; int j; double value;};
    struct COEF coef;

    while (!feof(fStrmA))  {
        fread((char *)&coef,sizeof(struct COEF),1,fStrmA);
        if (feof(fStrmA)) break;
        (*rowCount)[(coef.i)-1]++;        // increment rowCount for row i
    }
    fclose(fStrmA);
    return;
}


/**=====================================================================**/
void get_command_line_options(int argc, char **argv) {
// Process the command line options.

    if (getcfa("-A", argc, argv, Afilename))  {
    }

    if (getcfa("-b", argc, argv, bfilename))  {
    }

    if (getcfa("-solver", argc, argv, solvername))  {
    }

    if (getcfa("-pc", argc, argv, pcname))  {
    }

    if (getcfa("-?", argc, argv, Afilename))  {
        cout << "Usage: test [-A matrix_file] [-b rhs_file]" << endl;
        exit (1);
    }
    return;
}


/**=====================================================================**/
int getcfa(const char *flag, int argc, char **argv, char *argstr) {
/** getcfa - get command line flag argument
    Returns in "argstr" the argument following "flag". If the following
    argument begins with a dash or "flag" is the last token on the
    command line, "argstr" is returned as a null string **/

    int i;

    if (flag == 0 || argstr == 0)
        return (0); // flag or argstr is the NULL pointer

/* Process all command line arguments beginning with one past program name */
    for (i = 1; i < argc; i++) {
        if (argv[i] != 0)  { // check for NULL pointer
            if (strncmp(flag,argv[i],strlen(flag)) == 0) {  /* flag found */
                /* set argstr */
                if (i != argc - 1 && argv[i+1][0] != '-')
                    strcpy(argstr,argv[i+1]); /* argstr is not empty */
                else
                    argstr[0] = '\0';  /* argstr is empty */
                return (1);
            }
        }
    }
    return (0);  /* flag was not found on command line */
}

// Now include source for generic driver utilities.
#include "driverUtils.cc"

